/////////////////////////////////////////////////////////////////////////////
//
// Interleaved External Memory Interface for TimePilot
//
// Version : 1.10
//
// Copyright(c) 2004 Tatsuyuki Satoh , All rights reserved
//
// License:
//   This source is under the GPL license.
//
// Histry:
//   2004. 8. 3 Change Slice Timming 100ns to 150ns
//              Supported 120ns Middle Speed ROM.
//   2004. 7.21 1st Release
//
/////////////////////////////////////////////////////////////////////////////

/****************************************************************************
  Memory Map

 +---------+-----+-----------------------------------------+
 |ROM Area |Parts| Contents                                |
 +---------+-----+-----------------------------------------+
 |0000-1fff| H2? |MainCPU ROM 0000-1fff                    |
 |2000-3fff| H3? |MainCPU ROM 2000-3fff                    |
 |4000-5fff| H4? |MainCPU ROM 4000-5fff                    |
 |6000-7fff| H5? |MainCPU ROM 6000-7ffff(no use)           |
 +---------+-----+-----------------------------------------+
 |6000-60ff| E12 |Tilemap ColorLookUpTable ROM (*1)(*2)    |
 |6100-61ff| E9  |Object  ColorLookUpTable ROM (*1)(*2)    |
 |6200-621f| B4  |Palette ROM H                (*1)(*2)    |
 |6220-623f| B5  |Palette ROM L                (*1)(*2)    |
 |6240-625f|(B4) |YPrPb PaletteTable for HDTV  (*1)(*2)(*3)|
 |6260-627f|(B5) |YPrPb PaletteTable for HDTV  (*1)(*2)(*3)|
 +---------+-----+-----------------------------------------+
 |8000-9fff| C10?|Object Pattern ROM1                      |
 |a000-bfff| C11?|Object Pattern ROM2                      |
 +---------+-----+-----------------------------------------+
 |c000-dfff| F11 |Tilemap Pattern ROM                      |
 +---------+-----+-----------------------------------------+
 |e000-efff| A7  |SoundCPU ROM 0000-0fff                   |
 |f000-ffff|(A8) |not used                                 |
 +---------+-----+-----------------------------------------+
 *1: These overlap with MainCPU. 
 *2: load to internal memory after reset.
 *3: reserved for future.
****************************************************************************/

///////////////////////////////////////////////////////////////////////////

module extmem(
// Sync Clock
  CLK3x, PHASE3x, HCNT,
// Static Memory I/F
  MEM_A,MEM_DI,MEM_DO,MEM_WE,MEM_WO,MEM_OE,
// Main CPU
  MROM_A,MROM_DO,MROM_OE,
// Sub CPU
  SROM_A,SROM_DO,SROM_OE,
// Tilemap Pattern
  TROM_A,TROM_DO,
// Object Pattern
  OROM_A,OROM_DO,
// Configration Write
  CFG_EN,CFG_A,CFG_DO
);

/////////////////////////////////////////////////////////////////////////////
// module I/O assign
/////////////////////////////////////////////////////////////////////////////
input CLK3x;         // x3 pixel clock (18.432MHz)
input [2:0] PHASE3x; // x3 clock phase
input [1:0] HCNT;

output [15:0] MEM_A;
input  [7:0] MEM_DI;
output [7:0] MEM_DO;
output MEM_WE;
output MEM_WO;
output MEM_OE;

// Main CPU
input  [14:0] MROM_A;
output [7:0] MROM_DO;
input MROM_OE;

// Sub CPU
input  [12:0] SROM_A;
output [7:0] SROM_DO;
input SROM_OE;

// Tilemap Pattern
input  [12:0] TROM_A;
output [7:0] TROM_DO;

// Object Pattern
input  [13:0] OROM_A;
output [7:0] OROM_DO;

// Configration Write
input CFG_EN;
input [9:0] CFG_A;
output [7:0] CFG_DO;

/////////////////////////////////////////////////////////////////////////////

wire [15:0] cpu_a  = { 1'b0   , MROM_A };  //0000-7fff
wire [15:0] obj_a  = { 2'b10  , OROM_A };  //8000-bfff
wire [15:0] tile_a = { 3'b110 , TROM_A };  //c000-dfff
wire [15:0] sub_a  = { 3'b111 , SROM_A };  //e000-ffff
wire [15:0] cfg_a_ = { 6'b011000,CFG_A };  //6000-623f TCLUT+OCLUT+PAL1+PAL2

wire [15:0] co_a   = CFG_EN ? cfg_a_ : obj_a;

//
// interleave ROM
//         +--------+--------+--------+--------+--------+-----+
// HCNT    |   00   |   01   |   10   |   11   |   00   |  01 |
// PHASE3x |00|01|10|00|01|10|00|01|10|00|01|10|00|01|10|00|01|
// ROM_A   | OBJ | MAIN   |  TILE  |   MAIN |  OBJ   | MAIN   |
//         +--------------+--------+--------+--------+--------+
//
// ROM   Access speed: 150ns(162ns-@) or faster
//
// MAIN  Access cycle: 1.536MHz , synchronize in fetch of a M1 cycle. 
// SUB   Access cycle: 1.536MHz , worst case 651ns+163ns=814ns < 1123ns(*1)
// TILE  Access cycle: 1.536MHz , Do not roll over HCNT[1:0] phase
// OBJ   Access cycle: 1.536MHz , Do not roll over HCNT[1:0] phase
//
// *1:sub cpu clk == 1.78MHz == 562us / clock
//    M1 Access time are 2clock - (address setup from raise CLK) = 1123ns-@
//
// Can't change any multiplex order!
//

reg [15:0] rom_a;
reg [7:0] mrom_r , trom_r , orom_r , srom_r;

always @(posedge CLK3x)
begin
  if(PHASE3x[1])
  begin
    case(HCNT[1:0])
    2'b00:begin rom_a <= tile_a; srom_r <= MEM_DI; end
    2'b01:begin rom_a <= co_a;   trom_r <= MEM_DI; end
    2'b10:begin rom_a <= cpu_a;  orom_r <= MEM_DI; end
    2'b11:begin rom_a <= sub_a;  mrom_r <= MEM_DI; end
    endcase
  end
end

assign MEM_A  = rom_a;
assign MEM_DO = 8'h00;
assign MEM_WO = 1'b0;
assign MEM_WE = 1'b0;
assign MEM_OE = 1'b1;

assign OROM_DO = orom_r;
assign TROM_DO = trom_r;
assign MROM_DO = MROM_OE ? mrom_r : 8'h00;
assign SROM_DO = SROM_OE ? srom_r : 8'h00;

assign CFG_DO  = orom_r;

endmodule
