/////////////////////////////////////////////////////////////////////////////
//
// Internal Memory for TimePilot
//
// Version : 1.00
//
// Copyright(c) 2004 Tatsuyuki Satoh , All rights reserved
//
// License:
//   This source is under the GPL license.
//
// Histry:
//   2004. 8. 3 change data output async to sync
//              (for Xilinx BlockRAM auto assign)
//   2004. 7.23 1st Release
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// internal memory
/////////////////////////////////////////////////////////////////////////////

module imem_gen(
  rst_n,
//
  CLK_3x, PHASE_3x,
// SubCPU Memory
  SA,SDO,SDI,SCS,SWE,
// Tilemap RAM & OBJect RAM Write data
  V_DI,
// TileMap RAM
  TRAM_A,TRAM_DO,TRAM_CS,TRAM_WE,
// TileMap ColorLookUpTable ROM
  TCLUT_A,TCLUT_DO,

// Object RAM
  OBJ_A,OBJ0_DO,OBJ0_WE,OBJ1_DO,OBJ1_WE,
// Object Line RAM
  L0A,L0DO,L0DI,L0WE,
  L1A,L1DO,L1DI,L1WE,
// OBJ ColorLookUpTable ROM
  OCLUT_A,OCLUT_DO,

// Palette ROM
  PAL_A,PAL_EN,PAL_DO,

// Configration ROM write
  CFG_EN,CFG_A, CFG_DI
);

/////////////////////////////////////////////////////////////////////////////
// module I/O assign
/////////////////////////////////////////////////////////////////////////////
input rst_n;
input CLK_3x;       // x3 pixel clock
input [2:0] PHASE_3x; // x3 clock phase signal

// SubCPU Work RAM
input [9:0] SA;
output [7:0] SDO;
input [7:0] SDI;
input SCS;
input SWE;

// TileMap RAM & OBJect RAM Write data
input [7:0] V_DI;

// TileMap RAM
input [11:0] TRAM_A;
output [7:0] TRAM_DO;
input TRAM_CS,TRAM_WE;

// TileMap ColorLookUpTable ROM
input [7:0] TCLUT_A;
output [3:0] TCLUT_DO;

// Object RAM
input [7:0] OBJ_A;
output [7:0] OBJ0_DO,OBJ1_DO;
input OBJ0_WE,OBJ1_WE;

//Object ColorLookUpTable ROM
input [7:0] OCLUT_A;
output [3:0] OCLUT_DO;

// Object Line RAM
input [7:0] L0A,L1A;
input [3:0] L0DI,L1DI;
output [3:0] L0DO,L1DO;
input L0WE,L1WE;

// Palette ROM
input [4:0] PAL_A;
input PAL_EN;
output [14:0] PAL_DO;

// Configration Bus
output CFG_EN;
output [9:0] CFG_A;
input [7:0] CFG_DI;

/////////////////////////////////////////////////////////////////////////////
// Internal TCLUT & OCLUT PAL1 & PLA2 pre-load circuit
/////////////////////////////////////////////////////////////////////////////
reg [9:0] cfg_a_r; // 000-23f
reg [2:0] wcnt;    // 0.75MHz
wire cfg_busy = (cfg_a_r!=10'h240);

always @(posedge CLK_3x or negedge rst_n)
begin
  if(~rst_n)
  begin
    wcnt  <= 0;
    cfg_a_r <= 0;
  end else begin
    if(cfg_busy & PHASE_3x[2])
      {cfg_a_r,wcnt} <= {cfg_a_r,wcnt} + 1;
  end
end
assign CFG_EN = cfg_busy;
assign CFG_A  = cfg_a_r;

// address decode
wire tclut_we = cfg_busy & ~cfg_a_r[9] & ~cfg_a_r[8]; // 000-0ff TCLUT
wire oclut_we = cfg_busy & ~cfg_a_r[9] &  cfg_a_r[8]; // 100-1ff OCLUT
wire pal_we1 = cfg_busy & cfg_a_r[9] & ~cfg_a_r[5]; // 200-21f PAL H
wire pal_we2 = cfg_busy & cfg_a_r[9] &  cfg_a_r[5]; // 220-23f PLA L

/////////////////////////////////////////////////////////////////////////////
// Sub CPU SRAM
/////////////////////////////////////////////////////////////////////////////
reg [7:0] sub_ram[1023:0];
reg [7:0] sub_r;
always @(posedge CLK_3x)
begin
  if(SCS&SWE)  sub_ram[SA] <= SDI;
  sub_r <= sub_ram[SA];
end
assign SDO = SCS ? sub_r : 8'h00;

/////////////////////////////////////////////////////////////////////////////
// Video RAM
/////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////
// 1.Tilemap RAM
///////////////////////////////////////
reg [7:0] tram[4095:0];
reg [7:0] tram_r;
always @(posedge CLK_3x)
begin
  if(TRAM_CS&TRAM_WE) tram[TRAM_A] <= V_DI;
  tram_r <= tram[TRAM_A];
end
assign TRAM_DO = TRAM_CS ? tram_r : 8'h00;

///////////////////////////////////////
// 4.Object RAM x2
///////////////////////////////////////
reg [7:0] oram0[255:0],oram1[255:0];
reg [7:0] oram0_r,oram1_r;
always @(posedge CLK_3x)
begin
  if(OBJ0_WE) oram0[OBJ_A] <= V_DI;
  if(OBJ1_WE) oram1[OBJ_A] <= V_DI;
  oram0_r <= oram0[OBJ_A];
  oram1_r <= oram1[OBJ_A];
end
assign OBJ0_DO = oram0_r;
assign OBJ1_DO = oram1_r;

///////////////////////////////////////
// Object Line RAM x2
///////////////////////////////////////
reg [3:0] lram0[255:0],lram1[255:0];
reg [3:0] lram0_r,lram1_r;
always @(posedge CLK_3x)
begin
  if(L0WE) lram0[L0A] <= L0DI;
  if(L1WE) lram1[L1A] <= L1DI;
  lram0_r <= lram0[L0A];
  lram1_r <= lram1[L1A];
end
assign L0DO = lram0_r;
assign L1DO = lram1_r;

///////////////////////////////////////
// Tilemap ColorLookUpTable ROM(RAM)
// Object ColorLookUpTable ROM(RAM)
///////////////////////////////////////
wire [7:0] tc_a = cfg_busy ? cfg_a_r[7:0] : TCLUT_A;
wire [7:0] oc_a = cfg_busy ? cfg_a_r[7:0] : OCLUT_A;
reg [3:0] tc_ram[255:0],oc_ram[255:0];
reg [3:0] tc_r , oc_r;
always @(posedge CLK_3x)
begin
  if(tclut_we) tc_ram[tc_a] <= CFG_DI[3:0];
  if(oclut_we) oc_ram[oc_a] <= CFG_DI[3:0];
  tc_r <= tc_ram[tc_a];
  oc_r <= oc_ram[oc_a];
end
assign TCLUT_DO = tc_r;
assign OCLUT_DO = oc_r;

///////////////////////////////////////
// Palette ROM(RAM)
///////////////////////////////////////
reg [7:0] pal_ram0[31:0];
reg [6:0] pal_ram1[31:0];

wire [4:0] pal_ma = cfg_busy ? cfg_a_r[4:0] : PAL_A;

always @(posedge CLK_3x)
begin
  if(pal_we1) pal_ram0[pal_ma] <= CFG_DI;
  if(pal_we2) pal_ram1[pal_ma] <= CFG_DI[7:1];
end

// b = [14:10] , g = [ 9: 5] , r = [ 4: 0]
wire [14:0] pal_rom = {pal_ram0[pal_ma],pal_ram1[pal_ma]};
assign PAL_DO = pal_rom & {15{PAL_EN}};

endmodule
