/////////////////////////////////////////////////////////////////////////////
//
// TimePilot Video hardware
//
// Version : 1.00
//
// Copyright(c) 2004 Tatsuyuki Satoh , All rights reserved
//
// 1. License
//
// This program is freeware for non-commercial use. 
// The user of this source must assume all legal 
// responsibility in the use of this program. 
// It isn't suitable to put this source on GPL. 
//
// 2. Guarantee 
//
// An author does no guarantee about this program.
// You can use this under your own risk. 
// 
// 3. A rule to distribute.
//   Do either 1) or 2).
//   1)Include a perfect original source include this document.
//   2)Include all these headers , and add document of 
//     modification points
//
// Histry:
//   2004. 8.12 fix for support flip screen
//   2004. 8.10 Bug Fix on Altera Quartus2 v4.0sp1
//   2004. 7.21 1st Release
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
// Enternal Memory signal Assign.
//
// 1.Tilemap RAM
//
// Size    : 8bit x 2048word
// Address : TRAM_A[10:0]
// Din     : TRAM_DI[7:0]
// Dout    : none (from CPU)
// CS      : TRAM_CS
// WE      : TRAM_WE
// OE      : ~TRAM_WE
// note    : 2H interleave CPU and video scan
//
// 2.Tilemap pattern ROM
//
// Size    : 8bit x 8192word
// Address : TPAT_A[12:0]
// Din     : TPAT_DI[7:0]
// CS      : always 1
// OE      : always 1
// note    : 
//
// 3.Tilemap ColorLookupTable ROM
//
// Size    : 4bit x 256word (used only top 128word)
// Address : TCLUT_A
// Din     : TCLUT_DI[3:0]
// CS      : always 1
// OE      : always 1
// note    : 
//
// 4.Object RAM
//
// Size    : 8bit x 2port x 256word
// Address : ORAM_A[7:0]
// CS      : alaways 1
// Din1/2  : ORAM0_DI[7:0] / ORAM1_DI[7:0]
// Dout    : none(from CPU)
// WE1/2   : ORAM0_WE      / ORAM1_WE
// note    : 2H interleave CPU and video scan
//
// 5.Object pattern ROM
//
// Size    : 8bit x 16384word
// Address : OPAT_A[13:0]
// Din     : OPAT_DI[7:0]
// CS      : always 1
// OE      : always 1
// note    : 
//
// 6.Object ColorLookupTable ROM
//
// Size    : 4bit x 256word
// Address : OCLUT_A
// Din     : OCLUT_DI[3:0]
// CS      : always 1
// OE      : always 1
// note    : 
//
// 7.Object LINE RAM
//
// Size    : 4bit x 2line x 256word
// Address : LRAM0_A  / LRAM1_A
// Din     : LRAM0_DI / LRAM1_DI
// Dout    : LRAM0_DI / LRAM1_DI
// WE      : LRAM0_WE / LRAM1_WE
// CS      : always 1
// OE      : ~LRAM0_WE/ ~LRAM1_WE
// note    : access timming are cretical.
//
// 8.PaletteROM
//
// Size    : 15bit x 32word
// Address : PAL_A[4:0]
// Din     : none(connect to video dac)
// note    : asynchronous fast-access ROM
//
/////////////////////////////////////////////////////////////////////////////

module tp_vid(
// CPU interface
  A,di,do,n_rd,wait_n,
// address decoder chip select input
  tram_as,oram0_as,oram1_as,
// Scan Timming Input
  pclk,pclk2x,flip,
  hcnt    , vcnt,
  hwindow , vwindow,
// Tilemap RAM
  TRAM_A, TRAM_DI,TRAM_WE,TRAM_CS,
// Tilemap Pattern ROM
  TPAT_A,TPAT_DI,
// Tilemap ColorLookupTable ROM
  TCLUT_A,TCLUT_DI,
// Object RAM
  ORAM_A,ORAM0_DI,ORAM1_DI,ORAM0_WE,ORAM1_WE,
// Object pattern ROM
  OPAT_A,OPAT_DI,
// Object ColorLookupTable ROM
  OCLUT_A, OCLUT_DI,
// Object LINE RAM
  LRAM0_A,LRAM0_DO,LRAM0_DI,LRAM0_WE,
  LRAM1_A,LRAM1_DO,LRAM1_DI,LRAM1_WE,
// PaletteROM
  PAL_A,PAL_EN
);

// CPU interface
input  [11:0] A;
input  [7:0] di;
output [7:0] do;
input  n_rd;
output wait_n;

// address decoder chip select input
input tram_as;  //chip select tilemap RAM
input oram0_as; //chip select object RAM1
input oram1_as; //chip select object RAM2

// VIDEO input
input pclk;			// 6.144MHz video clock
input pclk2x;		// 12.288MHz x2 video clock
input flip;			// screen flip register
input [8:0] hcnt;	// H-counter
input [7:0] vcnt;	// V-counter
input hwindow;		// H-WINDOW
input vwindow;		// V-WINDOW

// Tilemap ram
output [11:0] TRAM_A;	// shared address , CPU & VIDEO & sound timming
input  [7:0] TRAM_DI;
output TRAM_WE,TRAM_CS;

// Tilemap Pattern ROM
output [12:0] TPAT_A;
input  [7:0] TPAT_DI;

// Tilemap ColorLookupTable ROM
output [7:0] TCLUT_A;
input  [3:0] TCLUT_DI;

// Object RAM
output [7:0] ORAM_A;
input  [7:0] ORAM0_DI; // Y / CODE
input  [7:0] ORAM1_DI; // X / color,flip
output ORAM0_WE,ORAM1_WE;

// Object pattern ROM
output [13:0] OPAT_A;
input  [7:0] OPAT_DI;

// Object ColorLookupTable ROM
output [7:0] OCLUT_A;
input  [3:0] OCLUT_DI;

// Object LINE RAM
output [7:0] LRAM0_A,LRAM1_A;
output [3:0] LRAM0_DO,LRAM1_DO;
input  [3:0] LRAM0_DI,LRAM1_DI;
output LRAM0_WE,LRAM1_WE;

// PaletteROM
output [4:0] PAL_A;
output PAL_EN;

// divided clk event for sync logic
wire raise_2h = (hcnt[1:0]==2'b01);
wire fall_2h  = (hcnt[1:0]==2'b11);

///////////////////////////////////////////////
// SCAN/CPU multiplexed tile ram address
///////////////////////////////////////////////

// flipped address
wire [7:3] hcp = hcnt[7:3] ^ {5{flip}};
wire [7:3] vcp = vcnt[7:3] ^ {5{flip}};
wire busy;

// tileram address
wire [11:0] tile_a = { hcnt[2],vcp[7:3],hcp[7:3] };

// multiplex bus
assign busy = hcnt[1];
//assign busy = hcnt[1] & hwindow & vwindow;
assign TRAM_A = busy ? tile_a : A;

//
// CPU write enable
//
wire vram_we = n_rd & (~hcnt[1]);

//
// TRAM / ORAM WE signal
//
assign TRAM_WE  = tram_as  & vram_we;
assign ORAM0_WE = oram0_as & vram_we;
assign ORAM1_WE = oram1_as & vram_we;
assign TRAM_CS  = 1'b1;

/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////
// CPU read bus controll
/////////////////////////////////////////////
reg  [7:0] cpu_do;
always @(posedge pclk)
begin
  if(raise_2h)
    cpu_do <= oram0_as ? ORAM0_DI :
              oram1_as ? ORAM1_DI :
                         TRAM_DI  ;
end

wire    video_as = (tram_as | oram0_as | oram1_as);  // 2004.8.10 Bug Fix
wire    video_oe = video_as & ~n_rd;                 // 2004.8.10 Bug Fix

assign do = video_oe ? cpu_do : 8'h00;
assign wait_n =~(video_as & hcnt[1] & ~n_rd);

/////////////////////////////////////////////
// Tilemap ram scan access
/////////////////////////////////////////////
reg [7:0] tile0;
reg [8:0] t_code;
reg [4:0] t_col , t_col_p;
reg th_flip  , tv_flip;
reg hwindow_pd; // pixlel data delayed hwindow

wire [7:0] tile1 = TRAM_DI;

always @(posedge pclk)
begin
  if(hcnt[1:0]==2'b11)
  begin

    if(~hcnt[2])
    begin
      // read tilemap code
      tile0  <= TRAM_DI;

      // tcol sync latch
      t_col   <= t_col_p;

      // display window skew
      if(hcnt[3]) // fall 8h
        hwindow_pd <= hwindow;
    end

    if(hcnt[2])
    begin
      // read color,flip
      tv_flip <=  tile0[7] ^ flip;
      th_flip <=  tile0[6] ^ flip;
      t_code  <= {tile0[5],tile1};
      t_col_p <= tile0[4:0];
    end
  end
end

/////////////////////////////////////////////
// Tilemap pattern ROM
/////////////////////////////////////////////
assign TPAT_A[12:4] = t_code;                   // tile code
assign TPAT_A[3]    = hcnt[2]   ^ th_flip;      // 4H byte select
assign TPAT_A[2:0]  = vcnt[7:0] ^ {3{tv_flip}}; // 4V,2V,1V

/////////////////////////////////////////////
// Tilemap pixel selector
/////////////////////////////////////////////
wire [1:0] t_pixel;

pixel_shift_4_1 t_pixel0
(.clk(pclk),.load(fall_2h),.dir(th_flip),.di(TPAT_DI[7:4]),.do(t_pixel[0]));
pixel_shift_4_1 t_pixel1
(.clk(pclk),.load(fall_2h),.dir(th_flip),.di(TPAT_DI[3:0]),.do(t_pixel[1]));

/////////////////////////////////////////////
// Tilemap ColorLookupTable ROM
/////////////////////////////////////////////
assign TCLUT_A = {1'b0,t_col,t_pixel};

/////////////////////////////////////////////////////////////////////////////
// Object(Sprite) controll
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////
// Object RAM
/////////////////////////////////////////////
wire hcnt8x = hcnt[7] ^ hwindow;

wire [7:0] obj_a = {2'b00,hcnt[7],hcnt8x,hcnt[6:4],hcnt[2]};
assign ORAM_A = busy ? obj_a : A[7:0];

/////////////////////////////////////////////
// OBJ ram scan access
/////////////////////////////////////////////

reg ov_match , ov_match_p;
reg [3:0] ov_line;
reg [7:0] oh_pos0 , oh_pos1 ,oh_pos_p;
reg [7:0] o_code;
reg [5:0] o_col , o_col_p;
reg oh_flip , oh_flip_p;
reg ov_flip , ov_flip_p;
reg lpage;

wire [7:0] ov_add = ORAM1_DI + vcnt[7:0];

always @(posedge pclk)
begin
  oh_pos0 <= oh_pos0 + 1;
  oh_pos1 <= oh_pos1 + 1;

  if(hcnt[1:0]==2'b11)
  begin

    //      rom , draw
    // 00 : [2] , [1] get h,color,flip
    // 01 : [3] , [2] get v,cole,change pat_rom access
    // 10 : [0] , [1] change color,match,draw page
    // 11 : [1] , [0]
    //

    // phase 0(2) : get color,flip,codeh
    if(~hcnt[2])
    begin
      // ORAM0 : Hpos 
      oh_pos_p  <= ORAM0_DI;
      // ORAM1 : color,flip
      ov_flip_p <=  ORAM1_DI[7];
      oh_flip_p <= ~ORAM1_DI[6];
      o_col_p   <=  ORAM1_DI[5:0];
    end

    // phase 1: switch pattern access
    if(hcnt[3:2]==2'b01)
    begin
      // ORAM0 : code
      o_code  <= ORAM0_DI;
      // ORAM1 : Vpos
      ov_match_p <= (ov_add[7:4]==4'b1111);
      ov_line    <= ov_add[3:0];
      // update OBJ info
      oh_flip <= oh_flip_p;
      ov_flip <= ov_flip_p;
   end

    // phase 2: switch OBJ pixel draw
   if(hcnt[3:2]==2'b10)
   begin
     // color , hpos , vmatch update
     o_col    <= o_col_p;
     ov_match <= ov_match_p;
     if(vcnt[0]) oh_pos1  <= oh_pos_p;
     else        oh_pos0  <= oh_pos_p;
     // reset draw position
     if(lpage^vcnt[0])
     begin
       if(vcnt[0]) oh_pos0 <= 0;
       else        oh_pos1 <= 0;
     end
     // flip lram page
     lpage <= vcnt[0];
   end
  end
end

/////////////////////////////////////////////
// Object pattern ROM
/////////////////////////////////////////////
assign OPAT_A[13:7] = o_code[7:1];
assign OPAT_A[6]    = o_code[0];
assign OPAT_A[5]    = ov_line[3]   ^ ov_flip;
assign OPAT_A[4:3]  = hcnt[3:2]    ^ {~oh_flip,oh_flip};
assign OPAT_A[2:0]  = ov_line[2:0] ^ {3{ov_flip}};

/////////////////////////////////////////////
// OBJ pixel make
/////////////////////////////////////////////
wire [1:0] o_pixel;

pixel_shift_4_1 o_pixel0
(.clk(pclk),.load(fall_2h),.dir(oh_flip),.di(OPAT_DI[7:4]),.do(o_pixel[0]));
pixel_shift_4_1 o_pixel1
(.clk(pclk),.load(fall_2h),.dir(oh_flip),.di(OPAT_DI[3:0]),.do(o_pixel[1]));

/////////////////////////////////////////////
// Object ColorLookupTable ROM
/////////////////////////////////////////////
wire OCLUT_A = { o_col , o_pixel};

/////////////////////////////////////////////
// Object LINE RAM
/////////////////////////////////////////////
reg [3:0] lram_di;
reg l_transparent;
wire o_transparent = (lram_di == 4'h0) | ~hwindow_pd;
wire [3:0] draw_d = OCLUT_DI;

// write enable
wire lclr_w  = (~pclk);
wire ldraw_w = (~pclk) & ov_match & l_transparent;

assign LRAM0_A  =  oh_pos0;
assign LRAM0_DO =  lpage ? 4'h0      : draw_d;
assign LRAM0_WE =  lpage ? lclr_w    : ldraw_w;
assign LRAM1_A  =  oh_pos1;
assign LRAM1_DO = ~lpage ? 4'h0      : draw_d;
assign LRAM1_WE = ~lpage ? lclr_w    : ldraw_w;

// get lram data
always @(negedge pclk)
begin
  // disp side
  lram_di <= lpage ? LRAM0_DI: LRAM1_DI;
  // draw side
  l_transparent <= (~lpage ? LRAM0_DI: LRAM1_DI) ==4'h0;
end

/////////////////////////////////////////////////////////////////////////////
// Final Output
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////
// Tilemap / Object Priority Multiplexer
/////////////////////////////////////////////
wire tile_sel = o_transparent | t_col[4];
wire [3:0] clut_mux = tile_sel ? TCLUT_DI : lram_di;

/////////////////////////////////////////////
// Palette ROM
/////////////////////////////////////////////
reg pal_window;

always @(posedge pclk2x)
begin
  pal_window <= hwindow_pd & vwindow;
end

assign PAL_A  = {tile_sel,clut_mux};
assign PAL_EN = pal_window;

endmodule

/////////////////////////////////////////////////////////////////////////////
// SubModule
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////
// pixel shift register module
// 4bit to 1bit x 4clk with flip select
/////////////////////////////////////////////////////
module pixel_shift_4_1(clk,load,dir,di,do);
input clk , load , dir;
input [3:0] di;
output do;

reg [3:0] pixel_r;

always @(posedge clk)
begin
  if( load ) pixel_r <= dir ? di : {di[0],di[1],di[2],di[3]};
  else   pixel_r[2:0] <= pixel_r[3:1];
end
wire do = pixel_r[0];

endmodule

