/////////////////////////////////////////////////////////////////////////////
//
// TimePilot Video Timming Generator
//
// Version : 1.00
//
// Copyright(c) 2004 Tatsuyuki Satoh , All rights reserved
//
// 1. License
//
// This program is freeware for non-commercial use. 
// The user of this source must assume all legal 
// responsibility in the use of this program. 
// It isn't suitable to put this source on GPL. 
//
// 2. Guarantee 
//
// An author does no guarantee about this program.
// You can use this under your own risk. 
// 
// 3. A rule to distribute.
//   Do either 1) or 2).
//   1)Include a perfect original source include this document.
//   2)Include all these headers , and add document of 
//     modification points
//
// Histry:
//   2004. 7.21 1st Release
//
/////////////////////////////////////////////////////////////////////////////

//////////////////////////////////
// 256(H) x 224(V) Display Window
// 384(H) x 262(V) Scan Timming
//////////////////////////////////

module tp_vtmg(
  rst_n,
  pclk, 
  hcnt, vcnt,
  hsync, vsync,
  hwindow , vwindow
);

// input
input rst_n; // reset input
input pclk;  // 6.144MHz Video Pixel Clock
// output
output [8:0] hcnt;  // H counter
output [7:0] vcnt;  // V counter
output hsync;       // H SYNC
output vsync;	    // V SYNC
output hwindow;	    // H Window (==~hblank)
output vwindow;	    // V Window (==~vblank)

/////////////////////////////////////////////////////////////////////////////
reg [8:0] hcnt_r;
reg hsync_r;

reg [7:0] vcnt_r;
reg vsync_r;
reg vwindow_r;

///////////////////////////////////////
// H timming generator
///////////////////////////////////////

//         +-------------------+--------+-------------------+
// hsync   |         0         |   1    |         0         |
// counter |000...0ff|180...19f|1a0..1bf|1c0...1ff|000...0ff|
// hwindow |    1    |             0              |    1    |
//         +---------+----------------------------+---------+

always @(posedge pclk or negedge rst_n)
begin
  if(~rst_n)
  begin
    hcnt_r <= 0;
    hsync_r <= 0;
  end else begin
    //hsync
    if( hcnt[8] && (hcnt_r[6:0] == 7'h20) )
      hsync_r <= 1;
    if(hcnt_r[4:0] == 5'h1f)
      hsync_r <= 0;
  
    //H counter
    hcnt_r <= (hcnt_r==255) ? 384 : hcnt_r + 1;

  end
end

///////////////////////////////////////
// V timming generator
///////////////////////////////////////

//         +------------------+--------+------------------+
// vsync   |    0             |    1   |        0         |
// counter |010...0ef|0f0..0ff|000..007|000..00f|010...0ef|
// vwindow |    1    |             0            |    1    |
//         +---------+--------------------------+---------+

always @(posedge hsync_r or negedge rst_n)
begin
  if(~rst_n)
  begin
    vwindow_r <= 0;
  end else begin
    // V Window
    if(vcnt_r[3:0]==4'hf)
      vwindow_r <= ~(vcnt_r[7:5]==3'b111);

    // V Sync
    if(vcnt_r==0)
      vsync_r <= ~vsync_r;
   end
end

// vcnt up
always @(negedge hcnt[8] or negedge rst_n)
begin
  if(~rst_n)
  begin
    vcnt_r <= 0;
  end else begin
    vcnt_r <= vcnt_r + 1;
    if(vsync_r)
      vcnt_r[3] <= 0;
  end
end

///////////////////////////////////////
// output assign
///////////////////////////////////////
assign hcnt    = hcnt_r;
assign vcnt    = vcnt_r;
assign hsync   = hsync_r;
assign vsync   = vsync_r;
assign hwindow = ~hcnt[8];
assign vwindow = vwindow_r;

endmodule
